# may be empty

from .gbid import PositionEncoder, VariantIdEncoder
from .client import (
    annotate_variants_list,
    annotate_variants_list_to_dataframe,
    parse_hgvs,
    annotate_dataframe_variants,
    whoami,
    lift_over_variants,
)
from .vcf_simple_annotator import annotate_vcf
from .json_simple_annotator import annotate_json
from .version import __version__



import logging
from itertools import islice
from .client import annotate_variants_list
from typing import Optional
from tqdm import tqdm


try:
    # try with a fast c-implementation ...
    import cyvcf2 as cyvcf2
except ImportError:
    # ... otherwise fallback python implementation
    logging.error(
        'No cyvcf2 module installed. Install, f.ex. by executing "pip install cyvcf2".'
    )


def annotate_vcf(
    input_vcf_path: str,
    output_vcf_path: str,
    genome: str = "hg38",
    use_ensembl: bool = True,
    use_refseq: bool = True,
    flatten_consequences: bool = True,
    batch_size: int = 500,
    username: Optional[str] = None,
    api_key: Optional[str] = None,
    use_netrc: bool = True,
    endpoint_url: str = "https://api.genebe.net/cloud/api-public/v1/variants",
    omit_acmg: bool = False,
    omit_csq: bool = False,
    omit_basic: bool = False,
    omit_advanced: bool = False,
    omit_normalization: bool = False,
    progress: bool = False,
):
    """
    Annotates a VCF file using the annotate_variants_list function.

    Args:
        input_vcf_path (str): Path to the input VCF file.
        output_vcf_path (str): Path to the output annotated VCF file.
        genome (str, optional): The genome version for annotation (e.g., 'hg38').
            Defaults to 'hg38'.
        use_ensembl (bool, optional): Whether to use Ensembl for annotation.
            Defaults to True.
        use_refseq (bool, optional): Whether to use RefSeq for annotation.
            Defaults to True.
        flatten_consequences (bool, optional): If set to False, return consequences as a list of objects.
            If set to True, only the most important consequence is returned in a flat form.
            Defaults to True.
        batch_size (int, optional): The size of each batch for processing variants.
            Defaults to 100. Must be smaller or equal 1000.
        username (str, optional): The username for authentication.
            Defaults to None.
        api_key (str, optional): The API key for authentication.
            Defaults to None.
        use_netrc (bool, optional): Whether to use credentials from the user's
            .netrc file for authentication. Defaults to True.
        endpoint_url (str, optional): The API endpoint for variant annotation.
            Defaults to 'https://api.genebe.net/cloud/api-public/v1/variants'.
        omit_acmg (bool, optional): Don't add ACMG scores in the output. Defaults to False.
        omit_csq (bool, optional): Don't add consequences in the output. Defaults to False.
        omit_basic (bool, optional): Don't add basic annotations (GnomAD frequencies etc) in the output. Defaults to False.
        omit_advanced (bool, optional): Don't add advanced annotations (ClinVar etc) in the output. Defaults to False.
        omit_normalization (bool, optional): Don't normalize variants. Use only if you are sure they are normalized already. Defaults to False.
        progress (bool, optional): Show progress bar
    """

    # Open VCF file
    vcf_reader = cyvcf2.VCF(input_vcf_path)

    # Add header to the VCF file
    vcf_reader.add_info_to_header(
        {
            "ID": "acmg_score",
            "Number": "1",
            "Type": "Integer",
            "Description": "ACMG Score by GeneBe",
        }
    )
    vcf_reader.add_info_to_header(
        {
            "ID": "acmg_criteria",
            "Number": ".",
            "Type": "String",
            "Description": "ACMG Score by GeneBe",
        }
    )
    vcf_reader.add_info_to_header(
        {
            "ID": "gnomad_exomes_af",
            "Number": "1",
            "Type": "Float",
            "Description": "GnomAD exomes Allele Frequency, by GeneBe",
        }
    )
    vcf_reader.add_info_to_header(
        {
            "ID": "gnomad_genomes_af",
            "Number": "1",
            "Type": "Float",
            "Description": "GnomAD genomes Allele Frequency, by GeneBe",
        }
    )

    # Open the output VCF file for writing
    vcf_writer = cyvcf2.Writer(output_vcf_path, vcf_reader)

    # Batch size for reading variants
    batch_size = 1000

    with tqdm() as pbar:
        # Iterate over variants in batches
        while True:
            # Read a batch of variants
            batch = list(islice(vcf_reader, batch_size))
            if not batch:
                break  # Break if there are no more variants

            # Process the batch
            variants_batch = [
                f"{variant.CHROM}-{variant.POS}-{variant.REF}-{variant.ALT[0]}"
                for variant in batch
            ]
            annotated_variants = annotate_variants_list(
                variants_batch,
                genome=genome,
                use_ensembl=use_ensembl,
                use_refseq=use_refseq,
                flatten_consequences=flatten_consequences,
                batch_size=batch_size,
                username=username,
                api_key=api_key,
                use_netrc=use_netrc,
                endpoint_url=endpoint_url,
                progress_bar=False,
                omit_acmg=omit_acmg,
                omit_csq=omit_csq,
                omit_basic=omit_basic,
                omit_advanced=omit_advanced,
                omit_normalization=omit_normalization,
            )

            # Add ACMG scores to the variants and write to the output VCF file
            for variant, annotated_variant in zip(batch, annotated_variants):
                fields_to_copy = [
                    "acmg_score",
                    "gnomad_exomes_af",
                    "gnomad_genomes_af",
                    "acmg_criteria",
#                    "consequences_ensembl",
#                    "consequences_refseq",
                ]

                for field in fields_to_copy:
                    value = annotated_variant.get(field)
                    if value is not None:
                        variant.INFO[field] = value

                vcf_writer.write_record(variant)
            pbar.update(len(batch))

    # Close the VCF writer
    vcf_writer.close()



import pandas as pd
import requests
from requests.auth import HTTPBasicAuth
from typing import Dict, List
import re
import json
import logging
import netrc
from tinynetrc import Netrc
from urllib.parse import urlparse
from importlib import metadata
from .version import __version__

try:
    from tqdm import tqdm
except ImportError:

    def tqdm(iterator, *args, **kwargs):
        return iterator


# Set up the User-Agent header
user_agent = f"GeneBe/{__version__} (requests)"
apikey_username_warning_showed = False


class TooManyRequestsError(Exception):
    pass


class WrongServerResponseError(Exception):
    pass


def _handle_too_many_requests():
    raise TooManyRequestsError(
        "Too many requests (429 error). Please provide a username and API key. "
        "More information at https://genebe.net/api"
    )


def _parse_variant_string(s):
    if len(s) > 0:
        s = s.strip()

    if len(s) == 0:
        return None

    pattern = re.compile(r"^[^-]+-\d+-[ACGT]*-[ACGT]*$")
    if pattern.match(s):
        chr_, pos, ref, alt = s.split("-")
        pos = int(pos)
        return {"chr": chr_, "pos": pos, "ref": ref, "alt": alt}
    else:
        print(f"Error: Invalid format in [{s}]")
        return None


def _save_credentials_to_netrc(machine, login, password):
    netrc = Netrc()
    netrc[machine]["login"] = login
    netrc[machine]["password"] = password
    netrc.save()


def _get_machine_name_from_endpoint(endpoint):
    """
    Extracts the machine name from an endpoint URL.

    Args:
        endpoint (str): The endpoint URL.

    Returns:
        str: The machine name extracted from the endpoint.
    """
    parsed_url = urlparse(endpoint)
    return parsed_url.netloc


def _read_netrc_credentials(endpoint):
    """
    Reads username and password from the .netrc file based on the machine name derived from the endpoint.

    Args:
        endpoint (str): The endpoint URL.

    Returns:
        tuple: A tuple containing (username, account, password). Returns (None, None, None) if no
        matching entry is found in the .netrc file.
    """
    machine_name = _get_machine_name_from_endpoint(endpoint)

    try:
        # Locate the .netrc file in the user's home directory
        netrc_file_path = netrc.netrc()

        # Get the authentication information for the specified machine
        login_info = netrc_file_path.authenticators(machine_name)

        if login_info:
            username, account, password = login_info
            return username, account, password
        else:
            return None, None, None

    except FileNotFoundError:
        print("Error: .netrc file not found.")
        return None, None, None


# Function to annotate variants
def annotate_variants_list(
    variants: List[str],
    genome: str = "hg38",
    use_ensembl: bool = True,
    use_refseq: bool = True,
    flatten_consequences: bool = True,
    batch_size: int = 500,
    username: str = None,
    api_key: str = None,
    use_netrc: bool = True,
    endpoint_url: str = "https://api.genebe.net/cloud/api-public/v1/variants",
    progress_bar: bool = True,
    omit_acmg: bool = False,
    omit_csq: bool = False,
    omit_basic: bool = False,
    omit_advanced: bool = False,
    omit_normalization: bool = False,
    annotator: str = "snpeff",
) -> List[Dict[str, object]]:
    """
    Annotates a list of genetic variants.

    Args:
        variants (List[str]): A list of genetic variants to be annotated.
            Format: chr-pos-ref-alt. Look at examples below.
        use_ensembl (bool, optional): Whether to use Ensembl for annotation.
            Defaults to True.
        use_refseq (bool, optional): Whether to use RefSeq for annotation.
            Defaults to True.
        genome (str, optional): The genome version for annotation (e.g., 'hg38').
            Defaults to 'hg38'.
        flatten_consequences (bool, optional): If set to False, return consequences as a list of objects.
            If set to True, only the most important consequence is returned in a flat form.
            Defaults to True.
        batch_size (int, optional): The size of each batch for processing variants.
            Defaults to 100. Must be smaller or equal 1000.
        username (str, optional): The username for authentication.
            Defaults to None.
        api_key (str, optional): The API key for authentication.
            Defaults to None.
        use_netrc (bool, optional): Whether to use credentials from the user's
            .netrc file for authentication. Defaults to True.
        endpoint_url (str, optional): The API endpoint for variant annotation.
            Defaults to 'https://api.genebe.net/cloud/api-public/v1/variants'.
        progress_bar (bool, optional): Show progress bar.
            Defaults to True.
        omit_acmg (bool, optional): Don't add ACMG scores in the output. Defaults to False.
        omit_csq (bool, optional): Don't add consequences in the output. Defaults to False.
        omit_basic (bool, optional): Don't add basic annotations (GnomAD frequencies etc) in the output. Defaults to False.
        omit_advanced (bool, optional): Don't add advanced annotations (ClinVar etc) in the output. Defaults to False.
        omit_normalization (bool, optional): Don't normalize variants. Use only if you are sure they are normalized already. Defaults to False.
        annotator (str, optional): Which VEP implementation to use.
            Defaults to snpeff.

    Returns:
        List[Dict[str, object]]: A list of dictionaries containing annotation information
        for each variant. The dictionary structure may vary. Check the current documentation
        on https://genebe.net/about/api

    Example:
        >>> variants = ["7-69599651-A-G", "6-160585140-T-G"]
        >>> annotations = annotate_variants_list(variants, use_ensembl=True,
        ...                                      use_refseq=False, genome='hg38',
        ...                                      batch_size=500, username="user123@example.com",
        ...                                      api_key="apikey456", use_netrc=False,
        ...                                      endpoint_url='https://api.genebe.net/cloud/api-public/v1/variants')
        >>> print(annotations)
        [{'chr': '7', 'pos':69599651 (...) }]

    Note:
        - The number of the elements in returned list is always equal to the number of queries.
    """

    if (use_refseq != True) and (use_ensembl != True):
        raise ValueError("use_refseq and use_ensembl cannot be both False")

    if (username is not None) and (api_key is not None):
        auth = (username, api_key)
        if use_netrc:
            _save_credentials_to_netrc(endpoint_url, username, api_key)
    elif use_netrc:
        username, _, api_key = _read_netrc_credentials(endpoint_url)
        if (username is not None) and (api_key is not None):
            auth = (username, api_key)
        else:
            auth = None
    else:
        auth = None

    # if username:
    if api_key == None:
        global apikey_username_warning_showed
        if not apikey_username_warning_showed:
            # show it only once
            logging.warning(
                f"You are not logged in to GeneBe. We recommend using username and api_key arguments. Find out more on https://genebe.net/api ."
            )
            apikey_username_warning_showed = True
    else:
        logging.warning(f"you will log in {username}")

    # input data validation
    # Convert the list of strings to list of dictionaries
    dict_list = [_parse_variant_string(s) for s in variants]
    logging.debug("I will query for " + json.dumps(dict_list))

    annotated_data = []

    ranges = range(0, len(dict_list), batch_size)
    iterator = tqdm(ranges) if progress_bar else ranges

    for i in iterator:
        # Prepare data for API request
        api_data = dict_list[i : i + batch_size]

        logging.debug("Querying for " + json.dumps(api_data))

        params = {"genome": genome}
        if use_refseq == False:
            params["useRefseq"] = use_refseq
        if use_ensembl == False:
            params["useEnsembl"] = use_ensembl
        if omit_acmg:
            params["omitAcmg"] = omit_acmg
        if omit_csq:
            params["omitCsq"] = omit_csq
        if omit_basic:
            params["omitBasic"] = omit_basic
        if omit_advanced:
            params["omitAdvanced"] = omit_advanced
        if omit_normalization:
            params["omitNormalization"] = omit_normalization
        if annotator:
            params["annotator"] = annotator

        # Make API request
        response = requests.post(
            endpoint_url,
            params=params,
            json=api_data,
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
                "User-Agent": user_agent,
            },
            auth=auth,
        )

        # Check if request was successful
        if response.status_code == 200:
            api_results_raw = response.json()
            api_results = [element for element in api_results_raw["variants"]]

            logging.debug("Backend result is " + json.dumps(api_results))

            # Append API results to annotated_data
            annotated_data.extend(api_results)
        elif response.status_code == 429:
            _handle_too_many_requests()
        else:
            logging.error(
                f"Got response with code {response.status_code} with body "
                + response.text
            )
            raise WrongServerResponseError(
                f"Got response with code {response.status_code} with body "
                + response.text
            )

    if flatten_consequences:
        for item in annotated_data:
            consequences_fields = None
            transcript = item.get("transcript")
            if transcript:  # check the ACMG chosen transcript
                if transcript.startswith("E"):
                    consequences_fields = "consequences_ensembl"
                else:
                    consequences_fields = "consequences_refseq"

                consequences_refseq = item[consequences_fields]
                if consequences_refseq:  # Check if the list is not empty
                    first_consequence = consequences_refseq[0]
                    for key in ["gene_symbol", "gene_hgnc_id", "transcript", "hgvs_c"]:
                        item[key] = first_consequence.get(key, None)
                    effects_list = first_consequence.get("consequences", None)
                    item["consequences"] = (
                        ",".join(
                            str(item) if item is not None else "None"
                            for item in effects_list
                        )
                        if effects_list is not None
                        else ""
                    )
            # Remove the 'consequences_*' fields, if exists
            item.pop("consequences_ensembl", None)
            item.pop("consequences_refseq", None)

    return annotated_data


def annotate_variants_list_to_dataframe(
    variants: List[str],
    use_ensembl: bool = True,
    use_refseq: bool = True,
    genome: str = "hg38",
    batch_size: int = 100,
    flatten_consequences: bool = True,
    username: str = None,
    api_key: str = None,
    use_netrc: bool = True,
    endpoint_url: str = "https://api.genebe.net/cloud/api-public/v1/variants",
) -> pd.DataFrame:
    # Call the existing function
    result_list = annotate_variants_list(
        variants=variants,
        use_ensembl=use_ensembl,
        use_refseq=use_refseq,
        flatten_consequences=flatten_consequences,
        genome=genome,
        batch_size=batch_size,
        username=username,
        api_key=api_key,
        use_netrc=use_netrc,
        endpoint_url=endpoint_url,
    )

    # Convert the list of dictionaries to a Pandas DataFrame
    df = pd.DataFrame(result_list)
    return df


class DnaChange:
    def __init__(self, chr, pos, ref, alt):
        self.chr = chr
        self.pos = int(pos)
        self.ref = ref
        self.alt = alt


def lift_over_variants(
    variants: List[str],
    from_genome: str = "hg19",
    dest_genome: str = "hg38",
    username: str = None,
    api_key: str = None,
    use_netrc: bool = True,
    endpoint_url: str = "https://api.genebe.net/cloud/api-public/v1/liftover",
) -> List[str]:
    """
    Lifts over a list of genetic variants between human genome versions.

    Args:
        variants (List[str]): A list of genetic variants represented as strings (chr-pos-ref-alt).
        from_genome (str): Source human genome version, one of hg19, hg38, t2t.
        dest_genome (str): Destination human genome version, one of hg19, hg38, t2t.
        username (str, optional): The username for authentication.
            Defaults to None.
        api_key (str, optional): The API key for authentication.
            Defaults to None.
        use_netrc (bool, optional): Whether to use credentials from the user's
            .netrc file for authentication. Defaults to True.
        endpoint_url (str, optional): The API endpoint.
            Defaults to 'https://api.genebe.net/cloud/api-public/v1/liftover'.



    Returns:
        List[str]: A list of lifted variants.

    Example:
        >>> input_variants = ['chr6-161006172-T-G']
        >>> from_genome = "hg19"
        >>> dest_genome = "hg38"
        >>> lifted_variants = lift_over_variants(input_variants, from_genome, dest_genome)
        >>> print(lifted_variants)
        ['chr6-160585140-T-G']
    """
    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "User-Agent": user_agent,
    }

    # Parse input strings into DnaChange objects
    dna_changes = [DnaChange(*variant.split("-")) for variant in variants]

    # Prepare the payload
    variants = [
        {
            "chr": dna_change.chr,
            "pos": dna_change.pos,
            "ref": dna_change.ref,
            "alt": dna_change.alt,
        }
        for dna_change in dna_changes
    ]

    print(f"this is my payload: {json.dumps(variants)}")

    if (username is not None) and (api_key is not None):
        auth = (username, api_key)
        if use_netrc:
            _save_credentials_to_netrc(endpoint_url, username, api_key)
    elif use_netrc:
        username, _, api_key = _read_netrc_credentials(endpoint_url)
        if (username is not None) and (api_key is not None):
            auth = (username, api_key)
        else:
            auth = None
    else:
        auth = None

    params = {
        "from": from_genome,
        "dest": dest_genome,
    }

    batch_size = 500
    # Make the POST request

    result = []
    for i in tqdm(range(0, len(variants), batch_size)):
        # Prepare data for API request
        chunk = variants[i : i + batch_size]
        response = requests.post(
            endpoint_url, params=params, json=chunk, headers=headers, auth=auth
        )

        # Check if the request was successful (status code 200)
        if response.status_code == 200:
            # Assuming the response contains a list of lifted variants
            lifted_variants = response.json()
            result.extend(
                (
                    f"{element.get('chr', '')}-{element.get('pos', '')}-{element.get('ref', '')}-{element.get('alt', '')}"
                    if element.get("chr")
                    else ""
                )
                for element in lifted_variants
            )
        else:
            # Print an error message if the request was not successful
            print(f"Error: {response.status_code}, {response.text}")
            return []
    return result


def parse_hgvs(
    hgvs: List[str],
    batch_size: int = 500,
    username: str = None,
    api_key: str = None,
    use_netrc: bool = True,
    endpoint_url: str = "https://api.genebe.net/cloud/api-public/v1/hgvs",
) -> List[str]:
    """
    Parses a list of genetic variants encoded in HGVS.

    Args:
        hgvs (List[str]): A list of genetic variants in hgvs format.
            Supports .n:, .c:, .g: and .m:. Look at examples below.
        batch_size (int, optional): The size of each batch for processing variants.
            Defaults to 500. Must be smaller or equal 1000.
        username (str, optional): The username for authentication.
            Defaults to None.
        api_key (str, optional): The API key for authentication.
            Defaults to None.
        use_netrc (bool, optional): Whether to use credentials from the user's
            .netrc file for authentication. Defaults to True.
        endpoint_url (str, optional): The API endpoint for parsing hgvs.
            Defaults to 'https://api.genebe.net/cloud/api-public/v1/hgvs'.

    Returns:
        List[str]: A list of dictionaries containing annotation information
        for each variant. Check the current documentation
        on https://genebe.net/about/api

    Example:
        >>> hgvs_variants = ["'NM_000277.2:c.1A>G"]
        >>> variants = parse_hgvs(hgvs_variants)
        >>> print(variants)
        ['12-102917129-AT-AC', '12-102852850-GA-G']

    Note:
        - The number of the elements in returned list is always equal to the number of queries.
        If some position cannot be parsed, an empty string is returned.
    """

    # logging in
    auth = None
    result = []

    for i in tqdm(range(0, len(hgvs), batch_size)):
        # Prepare data for API request
        chunk = hgvs[i : i + batch_size]

        logging.debug("Querying for " + json.dumps(chunk))
        # Make API request
        response = requests.post(
            endpoint_url,
            json=chunk,
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
                "User-Agent": user_agent,
            },
            auth=auth,
        )

        # Check if request was successful
        if response.status_code == 200:
            api_results_raw = response.json()
            logging.debug("Api reqult raw" + json.dumps(api_results_raw))
            api_results = [
                (
                    f"{element.get('chr', '')}-{element.get('pos', '')}-{element.get('ref', '')}-{element.get('alt', '')}"
                    if "chr" in element and "pos" in element
                    else ""
                )
                for element in api_results_raw
            ]

            logging.debug("Backend result is " + json.dumps(api_results))

            # Append API results to annotated_data
            result.extend(api_results)
        elif response.status_code == 429:
            _handle_too_many_requests()
        else:
            logging.error(
                f"Got response with code {response.status_code} with body "
                + response.text
            )
            raise WrongServerResponseError(
                f"Got response with code {response.status_code} with body "
                + response.text
            )
    return result


def annotate_dataframe_variants(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    Annotates genetic variants in a DataFrame using the 'annotate_variants_list_to_dataframe' function.

    Args:
        df (pd.DataFrame): Input DataFrame with genetic variant information. Must contain columns
        ["chr", "pos", "ref", "alt"]
        **kwargs: Additional keyword arguments to pass to 'annotate_variants_list_to_dataframe'.

    Returns:
        pd.DataFrame: A new DataFrame with additional annotation information.

    Example:
        >>> df = pd.DataFrame({'chr': ['6', '22'], 'pos': [160585140, 28695868], 'ref': ['T', 'AG'], 'alt': ['G', 'A']})
        >>> annotated_df = annotate_dataframe_variants(df, genome='hg38',use_ensembl=False,use_refseq=True, genome='hg38', flatten_consequences=True)
        >>> print(annotated_df)
    """
    # Ensure required columns are present in the DataFrame
    required_columns = ["chr", "pos", "ref", "alt"]
    missing_columns = [col for col in required_columns if col not in df.columns]
    if missing_columns:
        raise ValueError(f"Missing required columns in DataFrame: {missing_columns}")

    # Create a list of variant strings by concatenating columns
    variant_strings = [
        f"{row['chr']}-{row['pos']}-{row['ref']}-{row['alt']}"
        for _, row in df.iterrows()
    ]

    # Annotate variants using 'annotate_variants_list_to_dataframe'
    annotation_df = annotate_variants_list_to_dataframe(
        variants=variant_strings, **kwargs
    )
    annotation_df = annotation_df.drop(columns=["chr", "pos", "ref", "alt"])

    # Join the original DataFrame with the annotation results
    result_df = pd.concat([df.reset_index(drop=True), annotation_df], axis=1)

    return result_df


def whoami(
    username: str = None,
    api_key: str = None,
    use_netrc: bool = True,
    endpoint_url: str = "https://api.genebe.net/cloud/api-public/v1/whoami",
) -> List[str]:
    """
    Information about me: usage statistics, limits etc.

    Args:
        username (str, optional): The username for authentication.
            Defaults to None.
        api_key (str, optional): The API key for authentication.
            Defaults to None.
        use_netrc (bool, optional): Whether to use credentials from the user's
            .netrc file for authentication. Defaults to True.
        endpoint_url (str, optional): The API endpoint to use.
            Defaults to 'https://api.genebe.net/cloud/api-public/v1/whoami'.

    Returns:
        An object. Check the current documentation
        on https://genebe.net/about/api

    """

    # logging in
    if (username is not None) and (api_key is not None):
        auth = (username, api_key)
        if use_netrc:
            _save_credentials_to_netrc(endpoint_url, username, api_key)
    elif use_netrc:
        username, _, api_key = _read_netrc_credentials(endpoint_url)
        if (username is not None) and (api_key is not None):
            auth = (username, api_key)
        else:
            auth = None
    else:
        auth = None

    # Make API request
    response = requests.get(
        endpoint_url,
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": user_agent,
        },
        auth=auth,
    )

    # Check if request was successful
    if response.status_code == 200:
        data = response.json()
        return data
    else:
        logging.warning("Error:", response.status_code, response.text)
        return {}



try:
    # try with a fast c-implementation ...
    import mmh3 as mmh3
except ImportError:
    # ... otherwise fallback to this code!
    import pymmh3 as mmh3


class InvalidPositionException(Exception):
    def __init__(self, message):
        super().__init__(message)


class PositionEncoder:
    CHR_NOT_SUPPORTED = -1
    ERROR_WRONG_CHR_POSITION = -2

    def __init__(self):
        self.chromosomes = [
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10",
            "11",
            "12",
            "13",
            "14",
            "15",
            "16",
            "17",
            "18",
            "19",
            "20",
            "21",
            "22",
            "X",
            "Y",
            "M",
        ]

        self.chromosome_sizes = {
            "1": 248956422,
            "2": 242193529,
            "3": 198295559,
            "4": 190214555,
            "5": 181538259,
            "6": 170805979,
            "7": 159345973,
            "8": 145138636,
            "9": 138394717,
            "10": 133797422,
            "11": 135086622,
            "12": 133275309,
            "13": 114364328,
            "14": 107043718,
            "15": 101991189,
            "16": 90338345,
            "17": 83257441,
            "18": 80373285,
            "19": 58617616,
            "20": 64444167,
            "21": 46709983,
            "22": 50818468,
            "X": 156040895,
            "Y": 57227415,
            "M": 16569,
        }
        self.sizes = [
            self.chromosome_sizes.get(chr_name, 0) for chr_name in self.chromosomes
        ]
        self.offsets = [sum(self.sizes[:i]) for i in range(len(self.chromosomes))]

    def encode(self, chromosome: str, position: int):
        return self.encode_1_based_without_throw(self, chromosome, position)

    def encode_1_based_without_throw(self, chromosome, position):
        chr_name = self._chr(chromosome)
        try:
            chr_index = self.chromosomes.index(chr_name)
        except ValueError:
            # chromosome not supported
            return self.CHR_NOT_SUPPORTED
        else:
            size = self.sizes[chr_index]
            if position > size or position < 0:
                print(f"Wrong position for {chromosome}: {position}")
                return self.ERROR_WRONG_CHR_POSITION
            offset = self.offsets[chr_index]
            return position - 1 + offset

    @staticmethod
    def _chr(chromosome):
        # This method corresponds to ChromosomeNameNormalizer.chr(chromosome) in the Java code
        return chromosome  # You may need to implement your normalization logic here


class VariantIdEncoder:
    RADIX = 36
    MAX_DEL_LENGTH = (2**7) - 1
    MAX_INS_LENGTH = 9

    MASK_HASH = 0b1000000000000000000000000000000000000000000000000000000000000000
    MASK_NULL = 0b0100000000000000000000000000000000000000000000000000000000000000
    MASK_POS = 0b0011111111111111111111111111111111000000000000000000000000000000
    MASK_C_HASH = 0b0000000000000000000000000000000000100000000000000000000000000000
    MASK_INS_LEN = 0b0000000000000000000000000000000000011110000000000000000000000000
    MASK_DEL = 0b0000000000000000000000000000000000000001111111000000000000000000
    MASK_INS = 0b0000000000000000000000000000000000000000000000111111111111111111
    MASK_CHANGE_HASH_VALUE = (
        0b0000000000000000000000000000000000011111111111111111111111111111
    )
    BASES = ["A", "C", "G", "T"]

    positionEncoder = PositionEncoder()

    def set_value(self, input, mask, value):
        shift_count = (mask & -mask).bit_length() - 1
        new_value = (input & ~mask) | ((value << shift_count) & mask)
        return new_value

    def encode_bases(self, alt):
        bit_offset = 0
        encoded = 0
        for c in alt:
            value = 0
            if c == "A":
                value = 0
            elif c == "C":
                value = 1
            elif c == "G":
                value = 2
            elif c == "T":
                value = 3
            else:
                raise InvalidPositionException(
                    "Cannot encode alt as it contains strange letters: " + alt
                )

            encoded = self.set_value(encoded, 0b11 << bit_offset, value)

            bit_offset += 2

        return int(encoded)

    def hash(self, chromosome, position, del_length, alt):
        change = f"{chromosome}:{position}:{del_length}:{alt}"
        hash_value = mmh3.hash128(change.encode("utf-8"))
        value = self.set_value(hash_value, self.MASK_HASH, 1)
        return value

    def change_hash(self, del_length, alt):
        change = f"{del_length}:{alt}"
        hash_value = mmh3.hash128(change.encode("utf-8"))
        value = hash_value & self.MASK_CHANGE_HASH_VALUE
        return value

    def encode1based(self, chromosome, position, ref, alt):
        # make 0 based
        position = position - 1

        altNorm = alt.upper()

        while len(ref) > 0 and len(altNorm) > 0 and ref[0] == altNorm[0]:
            # strip leading common character, may happen in deletions / insertions
            position += 1
            ref = ref[1:]
            altNorm = altNorm[1:]

        # if ref is string than replace with length
        if isinstance(ref, str):
            ref = len(ref)

        refLength = ref

        chromosomeNormalized = chromosome.lstrip("chr").upper()
        positionId = self.positionEncoder.encode_1_based_without_throw(
            chromosomeNormalized, position
        )

        if positionId == -1:
            raise InvalidPositionException(
                "Wrong position: {}:{}.".format(chromosome, position)
            )

        if not set(altNorm).issubset("ACGNT"):
            raise InvalidPositionException(
                "Alt should contain only ACGNT, position: {}, altNorm: {}.".format(
                    position, altNorm
                )
            )

        useChangeHash = False

        if len(altNorm) > self.MAX_INS_LENGTH or refLength > self.MAX_DEL_LENGTH:
            useChangeHash = True

        encodedAlt = 0

        if "N" in alt:
            useChangeHash = True
        else:
            encodedAlt = self.encode_bases(altNorm)

        if positionId == -1:
            return hash(chromosomeNormalized, position, refLength, altNorm)
        else:
            if useChangeHash:
                encoded = 0
                encoded = self.set_value(encoded, self.MASK_HASH, 0)
                encoded = self.set_value(encoded, self.MASK_C_HASH, 1)
                encoded = self.set_value(encoded, self.MASK_POS, positionId)

                changeHash = self.change_hash(refLength, altNorm)
                encoded = self.set_value(
                    encoded, self.MASK_CHANGE_HASH_VALUE, changeHash
                )

                return encoded

            else:
                encoded = 0
                # print('1'+bin(encoded))
                encoded = self.set_value(encoded, self.MASK_HASH, 0)
                # print('2'+bin(encoded))
                encoded = self.set_value(encoded, self.MASK_C_HASH, 0)
                # print('3'+bin(encoded))
                encoded = self.set_value(encoded, self.MASK_POS, positionId)
                # print('4'+bin(encoded))
                encoded = self.set_value(encoded, self.MASK_INS_LEN, len(altNorm))
                # print('5'+bin(encoded))
                encoded = self.set_value(encoded, self.MASK_DEL, refLength)
                # print('6'+bin(encoded))
                encoded = self.set_value(encoded, self.MASK_INS, encodedAlt)
                # print('7'+bin(encoded))
                return encoded


# v = VariantIdEncoder()
# print(v.encode1based("1", 12, "A", "C"))



from setuptools import setup, find_packages
from genebe import version

setup(
    name="genebe",
    version=version.__version__,
    packages=find_packages(),
    install_requires=["pymmh3", "tinynetrc", "pandas", "requests", "tqdm", "jsonlines"],
    extras_require={
        "cpp": ["mmh3", "cyvcf2"],
    },
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Piotr Stawinski",
    description="GeneBe Client: A user-friendly system for annotating genetic variants",
    url="https://github.com/pstawinski/pygenebe",
    entry_points={
        "console_scripts": [
            "genebe=genebe.entrypoint:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
